# Copyright 2019-2020 Cambridge Quantum Computing
#
# Licensed under a Non-Commercial Use Software Licence (the "Licence");
# you may not use this file except in compliance with the Licence.
# You may obtain a copy of the Licence in the LICENCE file accompanying
# these documents or at:
#
#     https://cqcl.github.io/pytket/build/html/licence.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the Licence for the specific language governing permissions and
# limitations under the Licence, but note it is strictly for non-commercial use.

import numpy as np
from typing import Iterable, Tuple, Set
from pytket.circuit import Circuit, Bit, Qubit
from pytket.pauli import Pauli, QubitPauliString
from .operators import QubitPauliOperator


def append_pauli_measurement(pauli_string: QubitPauliString, circ: Circuit) -> None:
    """Appends measurement instructions to a given circuit, measuring each qubit in a given basis.

    :param pauli_string: The pauli string to measure
    :type pauli_string: QubitPauliString
    :param circ: Circuit to add measurement to.
    :type circ: Circuit
    """
    measured_qbs = []
    for qb, p in pauli_string.map.items():
        if p == Pauli.I:
            continue
        measured_qbs.append(qb)
        if p == Pauli.X:
            circ.H(qb)
        elif p == Pauli.Y:
            circ.Rx(0.5, qb)
    for b_idx, qb in enumerate(measured_qbs):
        unit = Bit(b_idx)
        circ.add_bit(unit, False)
        circ.Measure(qb, unit)


def _all_pauli_measurements(
    operator: QubitPauliOperator, circ: Circuit
) -> Iterable[Circuit]:
    """For each term in the operator, yields a copy of the given circuit with the appropriate measurements on each qubit. 
    The trivial term is omitted.

    :param operator: The operator
    :type operator: QubitPauliOperator
    :param circ: The circuit generating the desired state
    :type circ: Circuit
    :return: List of circuits in order of term from the operator
    :rtype: Iterable[Circuit]
    """
    for pauli_string in operator._dict.keys():
        if not pauli_string.map:
            continue
        copy = circ.copy()
        append_pauli_measurement(pauli_string, copy)
        yield copy
